import time as T
import math
import networkx as N
import sys
import MySQLdb

"""
computes connected components and hyperincident connected
components in social bookmarking datasets which are stored in a
database.

usage: python hcc.py [cc|hcc|hccplot|hccstats] db'

cc computes the connected components.
hcc computes the 2-incident connected components.
hccstats outputs basic statistics about the found components.
hccplot creates .plt-files of the component size distributions.


where db is the name of a database reachable through
dbconnection() and containing at least the tables and fields
defined in template.sql

Please cite:
Neubauer, N. and Obermayer, K. (2009a). Hyperincident connected
components of tagging networks. In Cattuto, C., Ruffo, G., and
Menczer, F., editors, HT '09: Proceedings of the 20th ACM
conference on Hypertext and hypermedia, pages 229-238, New York,
NY, USA.  ACM

For questions or comments, please contact
nicolas.neubauer@gmail.com

(c) Nicolas Neubauer 2008-2012
"""



def dbconnection(database):
    _host = "127.0.0.1"
    _user = ""
    _pwd = "" 

    db = MySQLdb.connect(host=_host, user=_user, passwd=_pwd, db=database)
    return db




def sql_chunk(sql, cursor, start, chunksize, add):
    cursor.execute(sql % (start, (chunksize+start if add else chunksize)))
    return cursor.fetchall()

def getstart(cursor, tablename, fieldname):
    if tablename=="" or fieldname=="":
        raise Exception("When no start is given to chunk_retrieve, please provide table- and id fieldname so we can find the starting id!")
    sql = "SELECT MIN(%s) FROM %s" % (fieldname, tablename)
    cursor.execute(sql)
    x = cursor.fetchone()
    return x[0]
    
def chunk_retrieve(sql, cursor, chunksize=1000, chunk_function=sql_chunk, add=False, start=-1, tablename="", fieldname=""):
    if start==-1:
        start = getstart(cursor, tablename, fieldname)
    while True:
        data = chunk_function(sql, cursor, start, chunksize, add)
        if not data: 
            return
        for element in data: 
            yield element  
        start += chunksize
        
class ConnectedComponents:

    def get_components(self, document_id, document_components, user_id,
                       user_components, tag_id, tag_components, max_component):
        if document_id in document_components:
            d = document_components[document_id]
        else:
            d = max_component+1
        
        if user_id in user_components:
            u = user_components[user_id]
        else:
            u = max_component+1
        
        if tag_id in tag_components:
            t = tag_components[tag_id]
        else:
            t = max_component+1
        return [d,u,t]

    def merge(self, a, b, cursor):
        if a==b:
            return
        cursor.execute("SELECT * FROM merge_components_copy WHERE component_1=%d AND component_2=%d" % (a,b))
        if cursor.fetchone():
            return
        cursor.execute("INSERT INTO merge_components_copy (component_1, component_2) VALUES (%d, %d)" % (a,b))

    def setc(self, id, component, cursor):
        cursor.execute("update entries set real_component = %d where id=%d" % (component, id))


    def singlerun(self, conn):
        """ first step: assign initial components, already performing trivial merges"""
        cursor = conn.cursor()
        cursor2 = conn.cursor()
        document_components = {}
        user_components = {}
        tag_components = {}
    
        max_component = 0
    
        cursor.execute("DELETE FROM merge_components_copy")
    
        d = chunk_retrieve("select id, document_id, user_id, tag_id from entries WHERE id>=%d AND id<%d ",
                                  cursor,
                                  chunksize=100000,
                                  add=True,
                                  tablename="entries",
                                  fieldname="id")
        
        first = True
        for (id, document_id, user_id, tag_id) in d:
            if first:
                first = False
                continue
            if id%1000==0:
                print id
            [d,u,t] = self.get_components(document_id, document_components, user_id, user_components, tag_id, tag_components, max_component)
            if d==max_component+1 and u==max_component+1 and t==max_component+1:
                max_component += 1
                document_components[document_id] = max_component
                user_components[user_id] = max_component
                tag_components[tag_id] = max_component
                self.setc(id, max_component, cursor2)
                continue
        
            if d<max_component+1:
                if u == max_component+1:
                    user_components[user_id] = d
                else:
                    self.merge(d, u, cursor2)
                if t == max_component+1:
                    tag_components[tag_id] = d
                else:
                    self.merge(d, t, cursor2)
                self.setc(id, d, cursor2)
                continue
            
            if u<max_component+1:
                document_components[document_id] = u
                if t == max_component+1:
                    tag_components[tag_id] = u
                else:
                    self.merge(d, t, cursor2)
                self.setc(id, u, cursor2)
                continue
        

            if t<max_component+1:
                document_components[document_id] = t
                user_components[user_id] = t
                self.setc(id, t, cursor2)            
                continue

            raise ProgrammingError("This should not happen.")



    def connect_components(self, conn):
        """second step: get all component_ids and replace them by a single one per connected "component component" """
        cursor = conn.cursor()
        cursor.execute("SELECT DISTINCT component_1 FROM merge_components_copy")
        component_ids = dict([(x[0],1) for x in cursor.fetchall()])
        print "read %d component ids in first position" % len(component_ids.keys())
        cursor.execute("SELECT DISTINCT component_2 FROM merge_components_copy")
        for x in cursor.fetchall():
            component_ids[x[0]] = 1
        component_ids = component_ids.keys()
        print "got %d component ids overall" % len(component_ids)
        G = N.Graph()
        for id in component_ids:
            G.add_node(id)
        cursor.execute("SELECT component_1, component_2 FROM merge_components_copy")
        i = 0
        for (c1, c2) in cursor.fetchall():
            G.add_edge(c1, c2)
            i += 1
            if i%1000==0:
                print "added %dth edge" % i
            
        print "finished edges: had %d edges" % i

        before = T.time() 
        components = N.connected_components(G)
        after = T.time()
        connection_time = after-before
        print "took %f seconds to create %d components from %d edges/%d nodes" % (connection_time, len(components), i, len(G.nodes()))
    
        before = T.time()
    
        for component in components:
            if len(component)==1:
                continue
            first_edge = component[0]
            other_edges = ""
            print "creating string" 
            for other_edge in component[1:]:
                other_edges = "%s%d," % (other_edges, other_edge)
            other_edges = other_edges[:-1]
            print "%d components -> component %d" % (len(component)-1, first_edge)
            sql = "UPDATE entries SET real_component=%s WHERE real_component in (%s)" % (first_edge, other_edges)
            #sqlfile.write(sql+"\n")
            cursor.execute(sql)
        after = T.time()
        update_time = after-before

    def stats(self, conn):
        cursor = conn.cursor()
        sql = "select real_component, count(*)  from entries group by real_component"
        cursor.execute(sql)
        results = cursor.fetchall()
        results = sorted(results, key=lambda item: -item[1])
        biggest_component = results[0][1]
        num_components = len(results)

        sql = "select count(*) FROM entries"
        cursor.execute(sql)
        num_edges = cursor.fetchone()[0]

        print "%d components, largest with %d edges, %f of all edges" % (num_components, num_edges, float(biggest_component)/num_edges)


        

    def compute_for(self, dbname):
        conn = dbconnection(dbname)    
        self.singlerun(conn)
        self.connect_components(conn)
        self.stats(conn)



class ComponentGetter():
    def __init__(self, idfield1, idfield2, idfield3):
        self.lastid1 = -1
        self.lastid2 = -1
        self.lastcomponent = -1
        self.idfield1 = idfield1
        self.idfield2 = idfield2
        self.idfield3 = idfield3
    
    def getComponent(self, id1, id2, new_component_id, cursor):
        if id1==self.lastid1 and id2==self.lastid2:
            return self.lastcomponent
        sql = "SELECT component FROM edge_components WHERE %s=%d AND %s=%d AND %s=-1" % (self.idfield1, id1, self.idfield2, id2, self.idfield3)
        cursor.execute(sql)
        result = cursor.fetchone()
        if result:
            component = result[0]
            self.lastcomponent = component
            self.lastid1 = id1
            self.lastid2 = id2
            return component
        return new_component_id
        
    def setComponent(self, id1, id2, component, cursor):
        sql = "INSERT INTO edge_components (%s,%s,%s,component) VALUES (%d,%d,-1,%d)" % (self.idfield1, self.idfield2, self.idfield3, id1, id2, component)
        cursor.execute(sql)
        self.lastid1 = id1
        self.lastid2 = id2
        self.lastcomponent = component

        
class HyperincidentConnectedComponents():

    def merge_components(self, c1, c2, cursor):
        cursor.execute("SELECT component_1 FROM merge_components_copy WHERE component_1=%d AND component_2=%d" % (c1, c2))
        a = cursor.fetchone()
        if a:
            return
        cursor.execute("INSERT INTO merge_components_copy (component_1, component_2) VALUES (%d, %d)" % (c1, c2))

    def append_or_merge(self, current_component, max_component_id, old_component1, getter1, id11, id12, old_component2, getter2, id21, id22, cursor):
        if not current_component==old_component1:
            if old_component1<=max_component_id:
                self.merge_components(current_component, old_component1, cursor)
            else:
                getter1.setComponent(id11, id12, current_component, cursor)

        if not current_component==old_component2:
            if old_component2<=max_component_id:
                self.merge_components(current_component, old_component2, cursor)
            else:
                getter2.setComponent(id21, id22, current_component, cursor)    


    def single_run_db(self, conn):
        print "starting single_run_db"
        cursor = conn.cursor()
        cursor2 = conn.cursor()
        max_component_id = 0
        cursor.execute("DELETE FROM merge_components_copy")
        cursor.execute("DELETE FROM edge_components")
    
    
        DU = ComponentGetter("document_id", "user_id", "tag_id")
        DT = ComponentGetter("document_id", "tag_id", "user_id")
        UT = ComponentGetter("user_id", "tag_id", "document_id")
    
        def get_components(document_id, user_id, tag_id, max_component_id, cursor):
            return [DU.getComponent(document_id, user_id, max_component_id+1, cursor),
                    DT.getComponent(document_id, tag_id, max_component_id+1, cursor),
                    UT.getComponent(user_id, tag_id, max_component_id+1, cursor)]
            
        sql = "SELECT id, document_id, user_id, tag_id FROM entries WHERE id>=%d AND id<%d ORDER BY id"
        d = chunk_retrieve(sql, cursor, chunksize=1000000, add=True, tablename="entries", fieldname="id")    
        for (id, document_id, user_id, tag_id) in d:
            if id%10000==0:
                print id
                                                    
            du = (document_id, user_id)
            dt = (document_id, tag_id)
            ut = (user_id, tag_id)
        
            components = get_components(document_id, user_id, tag_id, max_component_id, cursor2) 
        
            if min(components)==max_component_id+1:
                max_component_id += 1
                component_id = max_component_id
                DU.setComponent(document_id, user_id, component_id, cursor2)
                DT.setComponent(document_id, tag_id, component_id, cursor2)
                UT.setComponent(user_id, tag_id, component_id, cursor2)
                continue 
        
            # otherwise: one's already found
            if components[0]<=components[1] and components[0]<=components[2]:
                self.append_or_merge(components[0], max_component_id, components[1], DT, document_id, tag_id, 
                                     components[2], UT, user_id, tag_id, 
                                     cursor2)
                continue
        
            if components[1]<=components[0] and components[1]<=components[2]:
                self.append_or_merge(components[1], max_component_id, components[0], DU, document_id, user_id, 
                                     components[2], UT, user_id, tag_id, 
                                     cursor2)
                continue
        
            if components[2]<=components[0] and components[2]<=components[1]:
                self.append_or_merge(components[2], max_component_id, components[0], DU, document_id, user_id, 
                                     components[1], DT, document_id, tag_id,
                                     cursor2)
                continue
        
            # if we are here, we have a problem
            raise ProgrammingError("component ids not equal, but without minimum...? %s" % components)


        def merge_atonce(self, smaller_components, big_component, cursor):
            l = smaller_components.__str__().replace("]",")").replace("[","(").replace("L","")
            where = """ WHERE component_1 IN %s OR component_2 IN %s""" % (l, l)
    
            sql = """
            SELECT component_1, component_2  
            FROM merge_components_copy %s
            """ % where
    
            cursor.execute(sql)
            neighbours = {}
            for (c1, c2) in cursor.fetchall():
                if not c1 in smaller_components:
                    neighbours[c1] = 1
                    continue
                if not c2 in smaller_components:
                    neighbours[c2] = 1
                    continue
        
            print "found %d outside neighbours!" % len(neighbours)

            sql = "INSERT INTO merge_components_copy (component_1, component_2) VALUES (%s, %%s)" % big_component
            err = 0
            for c in neighbours: 
                try:
                    cursor.execute(sql, c) # executemany dies
                except Exception, e:
                    print e
                    err += 1
                    print err
                    continue
                

            sql = "UPDATE edge_components SET component=%d WHERE component in %s" % (big_component, l)
            print sql
            cursor.execute(sql)
    
            sql = "DELETE FROM merge_components_copy " + where
            cursor.execute(sql)


    def connect_components_pre(self, conn):
        """database version of connecting components. to be used until enough components have been merged that the rest
        can be solved much faster by NetworkX (which however causes memory problems with more than 7mio edges...)"""
        cursor = conn.cursor()
        sql = "SELECT COUNT(*) FROM merge_components_copy"
        cursor.execute(sql)
        x = cursor.fetchone()
        if (x[0]<6000000):
            print "no pre-connecting necessary!"
            return
    
    
        cursor2 = conn.cursor()
        sql = "SELECT MAX(component) FROM edge_components"
        cursor.execute(sql)
        next_id = cursor.fetchone()[0]+1


        ATONCE = 100
        MAXJOIN = 1000

        components = []
        
        n_changed = -1
        num_edges = 0    
        while n_changed!=0:
            n_changed = 0

            c = count_connecting_nodes(conn)
            components = c.keys()
            components.sort(cmp=lambda x,y: int(c[y]-c[x]))
            for key in components[0:10]:
                print key, c[key]
        
            a = components[0]
            print "choosing component %d, has %d neighbours" % (a, c[a])
                
            sql = "SELECT component_2 FROM merge_components_copy WHERE checked=0 AND component_1=%d" % a
            cursor.execute(sql)        
            b_components = dict([(x[0],1) for x in cursor.fetchall()])
            sql = "SELECT component_1 FROM merge_components_copy WHERE checked=0 AND component_2=%d" % a
            cursor.execute(sql)
            for (b2,) in cursor.fetchall():
                b_components[b2] = 1
            if len(b_components)==0:
                print "nothing found for %d" % b_components
                continue
        
            print "found %d neighbours" % len(b_components)
            dringesamt = 0
            drin = 0
            final_b = []
            for possible_b in components[1:]:
                if possible_b in b_components:
                    final_b.append(possible_b)
                    drin += 1
                    dringesamt = 0
                if drin>ATONCE:
                    cursor2.execute("SELECT COUNT(*) FROM merge_components_copy")
                    num_edges = cursor2.fetchone()[0]
                    print "num_edges: %d" %num_edges
                    print "merging!"
                    self.merge_atonce(final_b, a, cursor)

                    cursor2.execute("SELECT COUNT(*) FROM merge_components_copy")                
                    num_edges2 = cursor2.fetchone()[0]
                    print "num_edges: %d (%d)" % (num_edges2, num_edges2-num_edges)
                    if num_edges2<6000000:
                        return                

                
                    final_b = []
                    drin = 0
                if dringesamt>MAXJOIN:
                    print "now aborting, restarting"
                    break


    def connect_components(self, conn):
        """second step: get all component_ids and replace them by a single one per connected "component component" """
        cursor = conn.cursor()
        cursor.execute("SELECT DISTINCT component_1 FROM merge_components_copy")
        component_ids = dict([(x[0],1) for x in cursor.fetchall()])
        print "read %d component ids in first position" % len(component_ids.keys())
        cursor.execute("SELECT DISTINCT component_2 FROM merge_components_copy")
        for x in cursor.fetchall():
            component_ids[x[0]] = 1
        component_ids = component_ids.keys()
        print "got %d component ids overall" % len(component_ids)
        G = N.Graph()
        for id in component_ids:
            G.add_node(id)
        cursor.execute("SELECT component_1, component_2 FROM merge_components_copy")
        i = 0
        for (c1, c2) in cursor.fetchall():
            G.add_edge(c1, c2)
            i += 1
            if i%1000000==0:
                print "added %dth edge" % i
        print "finished edges: had %d edges" % i

        before = T.time() 
        components = N.connected_components(G)
        after = T.time()
        connection_time = after-before
        print "took %f seconds to create %d components from %d edges/%d nodes" % (connection_time, 
                                                                                  len(components), 
                                                                                  i, 
                                                                                  len(G.nodes()))
    
        before = T.time()
    
        ncomponents = 0
        for component in components:
            ncomponents += 1
            print "doing component %d (unifying %d components)" % (ncomponents, len(component))
            if len(component)==1:
                continue
            if ncomponents==1:
                print "setting all to giant component ;)"
                cursor.execute("UPDATE edge_components SET new_component=%d" % component[0])
                continue
            first_edge = component[0]
            other_edges = ""
            print "creating string" 
            i = 0
            for other_edge in component:
                # this looks ugly
                # but there may be SO many components involved that both executemany in WHERE ... in (...)
                # will eventually fail
                sql = "UPDATE edge_components SET new_component=%s WHERE component=%s" % (first_edge, other_edge)
                cursor.execute(sql)
                i+=1
                if (i%1000==0):
                    print "%d/%d: %d/%d" % (ncomponents, len(component), i, len(component))
        after = T.time()
        print "now updating all entries once..."
        cursor.execute("UPDATE edge_components SET component=new_component")
        update_time = after-before
    
    
        print "took %f seconds to update components in db" % update_time


    def set_edge_components(self, conn):
        cursor = conn.cursor()
        before = T.time()
        cursor.execute("UPDATE entries INNER JOIN edge_components ON entries.user_id = edge_components.user_id AND entries.tag_id = edge_components.tag_id AND -1 = edge_components.document_id SET entries.component = edge_components.component")
        after = T.time()
        print "SQL update query took %f seconds!" % (before-after)

    def get_component_ids(self, cursor):
        cursor.execute("SELECT id FROM hypercomponents")
        component_ids = [x[0] for x in cursor.fetchall()]
        return component_ids        


    def make_element_tables(self, conn):
        cursor = conn.cursor()
        component_ids = self.get_component_ids(cursor)
        for name in ['user', 'document', 'tag']:
            cursor.execute("delete from hypercomponents_%ss" % name)
            sql = """
            INSERT INTO hypercomponents_%(name)ss (component_id, %(name)s_id) 
            SELECT DISTINCT %(id)s, %(name)s_id FROM entries WHERE component=%(id)s"""
            i = 0
            for component_id in component_ids:
                if i%100==0:
                    print name, i, len(component_ids)
                i += 1
                dosql = sql % {'id': component_id, 'name': name}
                cursor.execute(dosql)

    def compute_component_properties(self, conn):
        cursor = conn.cursor()
        component_ids = self.get_component_ids(cursor)

  
        update_sql = """
        UPDATE hypercomponents 
        SET num_edges=%(edges)s,
            num_documents = %(document)s,
            num_tags = %(tag)s,
            num_users = %(user)s    
        WHERE
            id=%(id)s""" 
    
        for id in component_ids:
            data = {'id': id}
            for name in ['document', 'user', 'tag']:
                sql = "SELECT COUNT(*) FROM hypercomponents_%ss WHERE component_id=%d" % (name, id)
                cursor.execute(sql)
                data[name] = cursor.fetchone()[0]
            sql = "SELECT COUNT(*) FROM entries WHERE component=%d" % id
            cursor.execute(sql)
            data['edges'] = cursor.fetchone()[0]
            cursor.execute(update_sql % data)
                
            

    def make_components_table(self, conn):
        cursor  = conn.cursor()
        sql = "select distinct component from entries"
        cursor.execute(sql)
        ids = [x[0] for x in cursor.fetchall()]
        print "inserting %d ids" % len(ids)
        cursor.execute("DELETE FROM hypercomponents")
        sql = "INSERT INTO hypercomponents (id) VALUES (%s)"
        for id in ids:
            cursor.execute(sql, (id)) # "server goes away" during executemany ...

    def create_plots(self, dbname):
        conn = dbconnection(dbname)
        cursor = conn.cursor()
        for name in ['edge', 'document', 'user', 'tag']:
            print name
            cursor.execute("SELECT num_%ss, COUNT(*) FROM hypercomponents GROUP BY num_%ss" % (name,  name))
            h = {}
            for (count, num) in cursor.fetchall():
                c = h.get(count, 0)
                h[count] = c+num
            self.put_histogram(h, '%s_histo_%ss' % (dbname, name))


            cursor.execute("SELECT num_%ss FROM hypercomponents ORDER BY num_%ss DESC LIMIT 0,10" % (name, name))
            top10s = [r[0] for r in cursor.fetchall()]
            self.put_topN(top10s, '%s_top10_%ss' % (dbname, name))

    def put_topN(self, data, name):
        dataname = "%s.dat" % name
        f = open(dataname, "w")
        for d in data:
            f.write("%d\n" % d)
        f.close()

        plotname = "%s.plt" % name
        outname = "%s.mp" % name
        plt = """
set logscale y
set yrange[1:]
set xrange[-1:%d]
unset key
set term mp color 10.0
set out '%s'
set xlabel 'rank of component'
set ylabel 'size of component'
set style line 1 lt 1 lw 10
set style line 2 lt 2 lw 10
plot '%s' with imp ls 1""" % (len(data), outname, dataname)
        f = open(plotname, "w")
        f.write(plt)
        f.close()

        
            

    def put_histogram(self, h, name):
        import numpy
        manual_histogram = numpy.zeros((len(h.keys()), 2))
        i = 0
        for key in h.keys():
            manual_histogram[i,:] = (key, h[key])
            i += 1
        
        dataname = "%s.dat" % (name)
        pltname = "%s.plt" % (name)
        f = open(dataname, "w")
        for (x,y) in manual_histogram:
            f.write("%d %d\n" % (x,y))
        f.close()              


        plt = """set logscale x
        set logscale y
        unset key
        set term mp color 10.0
        set out '%s.mp'
        set xlabel 'component size'
        set ylabel 'number of components'
        plot '%s.dat'""" 

        f = open(pltname, "w")
        f.write(plt % (name, name))
        f.close()

        # cumulative
        n = sum(h.values())
        l = sorted(h.items(), key=lambda x: x[0])
        dataname = "%s_cumulative.dat" % name
        f = open(dataname, "w")
        for (size, count) in l:
            f.write("%d %d\n" % (size, n))
            n -= count
        f.close()

        
        pltname = "%s_cumulative.plt" % name
        f = open(pltname, "w")
        f.write(plt % ("%s_cumulative" % name, "%s_cumulative" % name))
        f.close()



        pltname = "%s_both.plt" % name
        plt = """set logscale x
        set logscale y
        unset key
        set term mp color 10.0
        set out '%s_both.mp'
        set xlabel 'component size'
        set ylabel 'number of components'
        plot '%s.dat' with dots, '%s_cumulative.dat' with points 0"""  % (name, name, name)
        f = open(pltname, "w")
        f.write(plt)
        f.close()

        

    def compute_for(self, dbname):
        conn = dbconnection(dbname)    
        self.single_run_db(conn)
        self.connect_components_pre(conn)
        self.connect_components(conn)
        self.set_edge_components(conn)
        self.make_components_table(conn)
        self.make_element_tables(conn)
        self.compute_component_properties(conn)

    def stats(self, dbname):
        conn = dbconnection(dbname)    
        cursor = conn.cursor()
        sql = "select num_edges from hypercomponents order by num_edges desc"
        cursor.execute(sql)
        results = cursor.fetchall()
        biggest_component = results[0][0]
        nlc = results[1][0]
        num_components = len(results)

        sql = "select count(*) FROM entries"
        cursor.execute(sql)
        num_edges = cursor.fetchone()[0]

        output = "%d components, largest with %d edges, %f of all edges, second-largest %d edges, %f of all edges, gcc/nlcc=%f"
        print output % (num_components,
                        biggest_component,
                        float(biggest_component)/num_edges,
                        nlc,
                        float(nlc)/num_edges,
                        float(nlc)/biggest_component)
        
        


if __name__=="__main__":
    action = sys.argv[1]
    db = sys.argv[2]
    if action=='cc':
        ConnectedComponents().compute_for(db)
    elif action=='hcc':
        HyperincidentConnectedComponents().compute_for(db)
    elif action=='hccplot':
        HyperincidentConnectedComponents().create_plots(db)
    elif action=='hccstats':
        HyperincidentConnectedComponents().stats(db)
    else:
        print 'usage: hcc.py [cc|hcc|hccplot|hccstats] db'
        print """cc computes the connected components.
hcc computes the 2-incident connected components.
hccstats outputs basic statistics about the found components.
hccplot creates .plt-files of the component size distributions.
"""
    
